/*

 path.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "path.h"
#include "partition.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "log.h"
#include "ideservd.h"
#include "directory.h"
#include "petscii.h"
#include "wchar.h"
#include "utf8.h"

void convertfilename(const Petscii *s, char typek, Petscii *name, Petscii *type, Petscii *mode) {
    int b, i, j;
    Petscii type2[4], mode2;
    int comma;

    i = comma = mode2 = 0; j = -1;

    for (b = 0; s[b] != 0; b++) {
        if (s[b] == ':') {
            i = comma = mode2 = 0; j = -1; continue;
        }
        if (s[b] == '/') {
            if (typek == 0) break;
        }

        switch (comma) {
        case 0:
            if (s[b] == typek) {
                comma++; j = 0;
                break;
            }
            if (s[b] == ',' || s[b] == '=') {
                comma = 3;
                break;
            }
            if (i < 16) {
                name[i++] = s[b];
            }
            break;
        case 1:
            if (s[b] == ',') {
                comma++;
                break;
            }
            if (s[b] == ',' || s[b] == '=') {
                comma = 3;
                break;
            }
            if (j < 3) {
                type2[j++] = s[b];
            }
            break;
        case 2:
            if (s[b] == ',' || s[b] == '=') {
                comma = 3;
                break;
            }
            mode2 = s[b];
            comma++; break;
        default:
            break;
        }
    }
    name[i] = 0;
    if (j == 1) {
        if (mode2 == 0) {
            switch (type2[0]) {
            case 'W':
            case 'R':
            case 'A':
            case 'M':
                mode2 = type2[0];
                if (mode2 == 'W' && type != NULL && type[0] == '*') {
                    type2[0] = 'S';
                    break;
                }
                j = -1;
                break;
            }
        }
    }
    if (j >= 0 && type != NULL) {
        if (j == 1) {
            switch (type2[0]) {
            case 'S': strcpy((char *)type2, "SEQ"); j = 3; break;
            case 'P': strcpy((char *)type2, "PRG"); j = 3; break;
            case 'U': strcpy((char *)type2, "USR"); j = 3; break;
            case 'B': strcpy((char *)type2, "DIR"); j = 3; break;
            case 'J': strcpy((char *)type2, "LNK"); j = 3; break;
            default: type2[0] = ':'; break;
            }
        }
        memcpy(type, type2, (unsigned int)j);
        type[j] = 0;
    }
    if (mode2 && mode) {
        *mode = mode2;
    }
}

int matchname(Petscii *a, Petscii *b) {
    while (*a == *b || (*b == '?' && *a != 0)) {
        if (*a == 0) {
            return 1;
        }
        a++; b++;
    }

    if (*b == '*') {
        b++;
        if (*b == 0) {
            return 1;
        }
        while (*a != 0) {
            if (matchname(a, b)) {
                return 1;
            }
            a++;
        }
    }

    return 0;
}

const Petscii *resolv_path(const Petscii *s, char *outpath, partition_t *outpart, Nameconversion nameconversion) {
    Petscii name[17];
    int fel = 0, fel2 = 0;
    partition_t partition = 0;
    size_t root;

    if (strchr((char *)s, ':')) {
        if (s[0] == '@') fel = 1;
        while ((s[fel] ^ 0x30) < 10) partition = partition * 10 + (s[fel++] ^ 0x30);
        while (s[fel] && s[fel] != '/' && s[fel] != ':') {
            fel++;
        }
        fel++;
    }
    {
        const char *path = partition_get_path(partition);
        if (path == NULL) {
            return NULL;
        }
        root = partition_get_root(partition);
        if (s[fel] == '/') {
            memcpy(outpath, path, root);
            outpath[root] = 0;
        } else {
            strcpy(outpath, path);
        }
    }

    for (;;) {
        Directory_entry direntry;
        Directory *directory;

        if (s[fel] == '/' || s[fel] == ':') {
            fel++; continue;
        }

        fel2 = fel;
        if (s[fel] != 0) {
            while (s[fel2] != 0 && s[fel2] != '/' && s[fel2] != ':') {
                fel2++;
            }
        }
        if (s[fel2] == 0) {
            if (outpart != NULL) *outpart = partition;
            return s + fel;
        }

        if (s[fel] == '.') {
            if (fel2 == fel + 1) {
                fel += 2; continue;
            }
            if (s[fel + 1] == '.' && fel2 == fel + 2) {
                char *o = strrchr(outpath, '/');
                if (o == NULL || o < outpath + root) o = outpath + root;
                *o = 0;
                fel += 3; continue;
            }
        }

        directory = directory_open(outpath, nameconversion, 0);
        if (directory == NULL) {
            errnotochannel15(errno);
            log_printf("Couldn't open the directory \"%s\": %s(%d)", outpath, strerror(errno), errno);
            return NULL;
        }
        convertfilename(s + fel, 0, name, NULL, NULL);

        while (directory_read(directory, &direntry))
        {
            if ((direntry.attrib & (A_DIR | A_CLOSED)) != (A_CLOSED | A_DIR)) {
                continue;
            }

            if (!matchname(direntry.name, name)) {
                continue;
            }

            if (strlen(directory_path(directory)) > 999) {
                direntry.attrib = 0;
            } else {
                strcpy(outpath, directory_path(directory));
            }
            break;
        }
        directory_close(directory);
        if (direntry.attrib == 0) return NULL;
        fel = fel2;
    }
}

static int badfilename(const char *name) {
    if (name[0] == '.') {
        if (name[1] == 0) return 1;
        if (name[1] == '.' && name[2] == 0) return 1;
    }
    return 0;
}

#ifdef _WIN32
static unsigned int petscii_to_utf8(Petscii c, char *d) {
    unsigned int w = petscii_to_unicode(c);
    if (w < 0x80) {
        *d = (char)w;
        return 1;
    }
    return utf8out(w, d);
}

void convertc64name(char *lname, const Petscii *name, const Petscii *type, Nameconversion nameconversion) {
    unsigned int i, f2 = 0;
    int doext;
    char dot;
    for (i = 0; name[i] != 0 && i < 16; i++) {
        f2 += petscii_to_utf8(name[i], lname + f2);
    }
    lname[f2] = 0;
    if (type == NULL) {
        return;
    }

    dot = nameconversion == NC_FORCEDOT ? '.' : ',';
    if (nameconversion == NC_IGNOREDOT) {
        doext = memcmp(type, "PRG", 3) != 0;
    } else {
        doext = type[0] != 0 || strchr((const char *)((lname[0] == '.') ? (lname + 1) : lname), '.') != NULL || badfilename(lname);
    }

    for (;;) {
        unsigned int f = f2;
        if (doext) {
            lname[f2++] = dot;
            for (i = 0; type[i] != 0 && i < 3; i++) {
                if (i == 1 && type[1] == 0xa0 && type[2] == 0xa0) break;
                f2 += petscii_to_utf8(type[i], lname + f2);
            }
            lname[f2] = 0;
        }
        if (!badfilename(lname)) break;
        doext = 1;
        f2 = f;
        dot = ',';
    }
}
#else
static size_t c64toascii(char *d, Petscii c, mbstate_t *ps) {
    ssize_t l;
    unsigned int w = petscii_to_unicode(c);
#if WCHAR_MAX <= 0xFFFFu
    if (w >= 0x10000) {
        l = wcrtomb(d, (wchar_t)(0xd800 + ((w - 0x10000) >> 10)), ps);
        if (l >= 0) {
            ssize_t l2 = wcrtomb(d + l, (wchar_t)(0xdc00 + (w & 0x3ff)), ps);
            l = l2 < 0 ? l2 : (l + l2);
        }
    } else
#endif
    if (w > '~') {
        l = (ssize_t)wcrtomb(d, (wchar_t)w, ps);
    } else {
        l = 1; *d = (char)w;
    }
    if (l >= 0) return (size_t)l;
#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
    l = sprintf(d, "?");
#else
    l = sprintf(d, w <= 0xffff ? "\\u%04x" : "\\U%08x", (unsigned int)w);
#endif
    return (l >= 0) ? (size_t)l : 0;
}

void convertc64name(char *lname, const Petscii *name, const Petscii *type, Nameconversion nameconversion) {
    mbstate_t ps;
    unsigned int i, f2 = 0;
    int doext;
    char dot;
    memset(&ps, 0, sizeof ps);
    for (i = 0; name[i] != 0 && i < 16; i++) {
        f2 += c64toascii(lname + f2, name[i], &ps);
    }
    lname[f2] = 0;
    if (type == NULL) {
        return;
    }

    dot = nameconversion == NC_FORCEDOT ? '.' : ',';
    if (nameconversion == NC_IGNOREDOT) {
        doext = memcmp(type, "PRG", 3) != 0;
    } else {
        doext = type[0] != 0 || strchr((lname[0] == '.') ? (lname + 1) : lname, '.') != NULL || badfilename(lname);
    }

    for (;;) {
        unsigned int f = f2;
        if (doext) {
            memset(&ps, 0, sizeof ps);
            lname[f2++] = dot;
            for (i = 0; type[i] != 0 && i < 3; i++) {
                if (i == 1 && type[1] == 0xa0 && type[2] == 0xa0) break;
                f2 += c64toascii(lname + f2, type[i], &ps);
            }
            lname[f2] = 0;
        }
        if (!badfilename(lname)) break;
        doext = 1;
        f2 = f;
        dot = ',';
    }
}
#endif

